"""Permission levels for the E27 Dealer API.

These values are not bitmask flags. The *_DISARMED variants are a panel-state
gate layered on top of the base permission level.
"""

from __future__ import annotations

from enum import Enum
from typing import Literal

from elke27_lib.errors import Elke27ProtocolError


class PermissionLevel(Enum):
    """Permission levels from the firmware, not a bitmask.

    *_DISARMED is a panel-state gate that requires all areas disarmed.

    Exact equivalent of PERMISSION_LEVEL_TYPE in the E27 firmware.
    Order and numeric values matter.
    """

    PLT_ENCRYPTION_KEY = 1
    PLT_ANY_USER = 2
    PLT_MASTER_USER = 3
    PLT_INSTALLER_USER = 4
    PLT_ENCRYPTION_KEY_DISARMED = 5
    PLT_ANY_USER_DISARMED = 6
    PLT_MASTER_USER_DISARMED = 7
    PLT_INSTALLER_USER_DISARMED = 8


_DISARMED_TO_BASE = {
    PermissionLevel.PLT_ENCRYPTION_KEY_DISARMED: PermissionLevel.PLT_ENCRYPTION_KEY,
    PermissionLevel.PLT_ANY_USER_DISARMED: PermissionLevel.PLT_ANY_USER,
    PermissionLevel.PLT_MASTER_USER_DISARMED: PermissionLevel.PLT_MASTER_USER,
    PermissionLevel.PLT_INSTALLER_USER_DISARMED: PermissionLevel.PLT_INSTALLER_USER,
}

AREA_GENERATORS = (
    ("area_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("area_get_def_flags", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("area_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_get_bypassed_zones", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_set_alarm_state", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_get_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_set_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_set_arm_state", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_alarm_test", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("area_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("area_num_not_rdy_zones", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_get_unsecure_zones", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("area_get_troubles", PermissionLevel.PLT_ENCRYPTION_KEY),
)

BARRIER_GENERATORS = (
    ("barrier_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("barrier_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("barrier_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("barrier_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("barrier_get_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("barrier_set_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("barrier_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

BUS_IOS_GENERATORS = (
    ("bus_ios_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("bus_ios_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("bus_ios_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("bus_ios_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("bus_ios_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("bus_ios_change_id", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("bus_ios_get_version_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("bus_ios_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("bus_ios_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("bus_ios_start_discovery", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("bus_ios_stop_discovery", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("bus_ios_get_error_counts", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("bus_ios_get_dev_error_counts", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("bus_ios_set_sn", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("bus_ios_fill_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

CELL_GENERATORS = (
    ("cell_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("cell_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("cell_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("cell_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
)

LOG_GENERATORS = (
    ("log_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("log_get_log", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("log_get_list", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("log_clear", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("log_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("log_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("log_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("log_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("log_get_index", PermissionLevel.PLT_ENCRYPTION_KEY),
)

CONTROL_GENERATORS = (
    ("control_authenticate", PermissionLevel.PLT_ANY_USER),
    ("control_get_version_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("control_get_ssp_version_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("control_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("control_get_table_alloc_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("control_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("control_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("control_silence_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("control_get_system_time", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("control_set_system_time", PermissionLevel.PLT_MASTER_USER),
    ("control_get_factory_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("control_set_factory_attribs", PermissionLevel.PLT_INSTALLER_USER),
    ("control_test_realloc", PermissionLevel.PLT_INSTALLER_USER),
    ("control_backup", PermissionLevel.PLT_INSTALLER_USER),
    ("control_restore", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("control_load_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

SYSTEM_GENERATORS = (
    ("system_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("system_get_system_time", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("system_set_system_time", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("system_set_system_key", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("system_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("system_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("system_get_cutoffs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("system_set_cutoffs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("system_get_sounders", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("system_reset_smokes", PermissionLevel.PLT_ENCRYPTION_KEY_DISARMED),
    ("system_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("system_get_troubles", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("system_r_u_alive", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("system_file_info", PermissionLevel.PLT_INSTALLER_USER),
    ("system_start_updt", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("system_set_run", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("system_reconfig", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("system_set_debug_flags", PermissionLevel.PLT_INSTALLER_USER),
    ("system_get_debug_flags", PermissionLevel.PLT_INSTALLER_USER),
    ("system_get_debug_string", PermissionLevel.PLT_INSTALLER_USER),
    ("template_get_configured", PermissionLevel.PLT_INSTALLER_USER),
    ("system_get_update", PermissionLevel.PLT_ENCRYPTION_KEY),
)

CS_PARAM_GENERATORS = (
    ("cs_param_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("cs_param_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("cs_param_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("cs_param_get_def_flags", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("cs_param_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("cs_param_get_cs_paths", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("cs_param_set_cs_paths", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("cs_param_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("cs_param_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("cs_param_commtest", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

FLAG_GENERATORS = (
    ("flag_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("flag_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("flag_set_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("flag_del_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("flag_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
)

KEYFOB_GENERATORS = (
    ("keyfob_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("keyfob_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("keyfob_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("keyfob_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("keyfob_change_id", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("keyfob_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("keyfob_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("keyfob_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
)

KEYPAD_GENERATORS = (
    ("keypad_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("keypad_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("keypad_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("keypad_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("keypad_change_id", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("keypad_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("keypad_get_version_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("keypad_get_def_flags", PermissionLevel.PLT_ENCRYPTION_KEY),
)

LIGHT_GENERATORS = (
    ("light_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("light_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("light_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("light_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("light_get_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("light_set_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("light_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

LOCK_GENERATORS = (
    ("lock_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("lock_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("lock_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("lock_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("lock_get_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("lock_set_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("lock_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("lock_get_user_code", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("lock_set_user_code", PermissionLevel.PLT_MASTER_USER_DISARMED),
)

NETWORK_PARAM_GENERATORS = (
    ("network_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("network_param_get_ssid", PermissionLevel.PLT_INSTALLER_USER),
    ("network_param_get_rssi", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("network_param_get_attribs", PermissionLevel.PLT_INSTALLER_USER),
    ("network_param_set_attribs", PermissionLevel.PLT_INSTALLER_USER),
    ("network_restart", PermissionLevel.PLT_INSTALLER_USER),
)

NET_DEV_GENERATORS = (
    ("link_function", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("logout_function", PermissionLevel.PLT_ANY_USER),
    ("net_dev_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("net_dev_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("net_dev_get_creds", PermissionLevel.PLT_INSTALLER_USER),
    ("net_dev_set_creds", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("net_dev_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("net_dev_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("net_dev_get_list", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("net_dev_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
)

OUTPUT_GENERATORS = (
    ("output_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("output_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("output_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("output_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("output_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("output_get_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("output_set_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("output_refresh_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("output_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("output_get_available", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("output_get_all_outputs_status", PermissionLevel.PLT_ENCRYPTION_KEY),
)

REGISTER_GENERATORS = (
    ("register_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("register_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("register_set_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("register_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("register_del_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
)

RULE_GENERATORS = (
    ("rule_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("rule_get_storage_size", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("rule_get_rules", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("rule_set_rules", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("rule_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("rule_process", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("rule_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
)

TASK_GENERATORS = (
    ("task_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("task_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("task_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("task_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("task_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("task_activate", PermissionLevel.PLT_ENCRYPTION_KEY),
)

TIMER_GENERATORS = (
    ("timer_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("timer_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("timer_set_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("timer_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("timer_del_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
)

TSTAT_GENERATORS = (
    ("tstat_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("tstat_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("tstat_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("tstat_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("tstat_get_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("tstat_set_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("tstat_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("tstat_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

USER_VALUE_GENERATORS = (
    ("user_value_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("user_value_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("user_value_set_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("user_value_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("user_value_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("user_value_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

USERGROUP_GENERATORS = (
    ("usergroup_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("usergroup_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("usergroup_set_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("usergroup_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("usergroup_del_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
)

USER_GENERATORS = (
    ("user_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("user_get_def_flags", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("user_get_attribs", PermissionLevel.PLT_ANY_USER),
    ("user_set_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("user_del_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("user_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("user_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

WIEGAND_GENERATORS = (
    ("wiegand_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("wiegand_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("wiegand_set_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("wiegand_del_attribs", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("wiegand_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("wiegand_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

WIN_COVER_GENERATORS = (
    ("win_cover_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("win_cover_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("win_cover_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("win_cover_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("win_cover_get_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("win_cover_set_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("win_cover_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("win_cover_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

WLTX_GENERATORS = (
    ("wltx_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("wltx_get_version_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("wltx_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("wltx_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("wltx_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("wltx_change_id", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("wltx_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("wltx_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("wltx_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("wltx_start_discovery", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("wltx_stop_discovery", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("wltx_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

ZONE_GENERATORS = (
    ("zone_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_get_defs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_get_def_flags", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zone_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_get_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_set_status", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zone_pwr_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zone_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zone_fill_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zone_get_available", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zone_get_all_zones_status", PermissionLevel.PLT_ENCRYPTION_KEY),
)

REPEATER_GENERATORS = (
    ("repeater_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("repeater_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("repeater_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("repeater_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("repeater_del_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("repeater_realloc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
)

ZWAVE_GENERATORS = (
    ("zwave_get_table_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_inclusion", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_exclusion", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_rem_failed", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_abort", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_restart", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("zwave_default", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_clear", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_get_attribs", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_set_attribs", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_get_configured", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_get_trouble", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_set_assoc", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_assign_return_route", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_heal_node", PermissionLevel.PLT_MASTER_USER_DISARMED),
    ("zwave_get_battery", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_set_scene", PermissionLevel.PLT_ANY_USER),
    ("zwave_set_config", PermissionLevel.PLT_INSTALLER_USER_DISARMED),
    ("zwave_get_rssi", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_get_mfr", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_get_route_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_get_node_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_get_secure_node_info", PermissionLevel.PLT_ENCRYPTION_KEY),
    ("zwave_get_supported_sensor", PermissionLevel.PLT_ENCRYPTION_KEY),
)


_GENERATOR_GROUPS = (
    AREA_GENERATORS,
    BARRIER_GENERATORS,
    BUS_IOS_GENERATORS,
    CELL_GENERATORS,
    LOG_GENERATORS,
    CONTROL_GENERATORS,
    SYSTEM_GENERATORS,
    CS_PARAM_GENERATORS,
    FLAG_GENERATORS,
    KEYFOB_GENERATORS,
    KEYPAD_GENERATORS,
    LIGHT_GENERATORS,
    LOCK_GENERATORS,
    NETWORK_PARAM_GENERATORS,
    NET_DEV_GENERATORS,
    OUTPUT_GENERATORS,
    REGISTER_GENERATORS,
    RULE_GENERATORS,
    TASK_GENERATORS,
    TIMER_GENERATORS,
    TSTAT_GENERATORS,
    USER_VALUE_GENERATORS,
    USERGROUP_GENERATORS,
    USER_GENERATORS,
    WIEGAND_GENERATORS,
    WIN_COVER_GENERATORS,
    WLTX_GENERATORS,
    ZONE_GENERATORS,
    REPEATER_GENERATORS,
    ZWAVE_GENERATORS,
)


def _build_generator_permission() -> dict[str, PermissionLevel]:
    permissions: dict[str, PermissionLevel] = {}
    for group in _GENERATOR_GROUPS:
        for generator_key, level in group:
            permissions[generator_key] = level
    return permissions


GENERATOR_PERMISSION = _build_generator_permission()
ALL_PERMISSION_KEYS = set(GENERATOR_PERMISSION.keys())


def canonical_generator_key(name: str) -> str:
    if name.startswith("generator_"):
        return name[len("generator_") :]
    return name


def permission_for_generator(generator_key: str) -> PermissionLevel:
    generator_key = canonical_generator_key(generator_key)
    try:
        return GENERATOR_PERMISSION[generator_key]
    except KeyError as exc:
        raise Elke27ProtocolError(
            f"No permission metadata for generator {generator_key!r}."
        ) from exc


def requires_pin(level: PermissionLevel) -> bool:
    """Return True when a PIN is required for this permission level."""

    return level not in {
        PermissionLevel.PLT_ENCRYPTION_KEY,
        PermissionLevel.PLT_ENCRYPTION_KEY_DISARMED,
    }


def requires_disarmed(level: PermissionLevel) -> bool:
    """Return True when *_DISARMED requires the panel to be disarmed."""

    return level.name.endswith("_DISARMED")


def strip_disarmed(level: PermissionLevel) -> PermissionLevel:
    """Return the base permission level for a *_DISARMED variant."""

    return _DISARMED_TO_BASE.get(level, level)


def required_role(
    level: PermissionLevel,
) -> Literal["encryption_key", "any_user", "master", "installer"]:
    """Return the minimum role required for the permission level."""

    base = strip_disarmed(level)
    if base is PermissionLevel.PLT_ENCRYPTION_KEY:
        return "encryption_key"
    if base is PermissionLevel.PLT_ANY_USER:
        return "any_user"
    if base is PermissionLevel.PLT_MASTER_USER:
        return "master"
    if base is PermissionLevel.PLT_INSTALLER_USER:
        return "installer"
    raise ValueError(f"Unknown permission level: {level}")



"""Network parameter request generators."""

from __future__ import annotations

ResponseKey = tuple[str, str]


def generator_network_param_get_ssid() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("network", "get_ssid")


def generator_network_param_get_rssi() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("network", "get_rssi")



"""
Live E27 test: system status snapshot via system_get_trouble.

Run:
  source ~/elk-e27-env-vars.sh
  pytest -q -m live_e27 test/test_live_e27_system_status.py -s
"""

from __future__ import annotations

from typing import cast

import pytest

from elke27_lib.client import Elke27Client
from elke27_lib.const import E27ErrorCode
from test.helpers.error_codes import describe_error, extract_error_code
from test.helpers.payload_validation import assert_payload_shape

_LIVE_TIMEOUT_S = 15.0


def _extract_error_code(payload: object) -> E27ErrorCode | None:
    if isinstance(payload, dict):
        payload_map = cast(dict[str, object], payload)
        direct = payload_map.get("error_code")
        if isinstance(direct, int):
            try:
                return E27ErrorCode(direct)
            except ValueError:
                return None
        for key in ("get_trouble", "get_troubles"):
            nested = payload_map.get(key)
            if isinstance(nested, dict):
                nested_map = cast(dict[str, object], nested)
                nested_code = nested_map.get("error_code")
                if isinstance(nested_code, int):
                    try:
                        return E27ErrorCode(nested_code)
                    except ValueError:
                        return None
    return None


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_system_status_snapshot(live_e27_client: Elke27Client) -> None:
    result = await live_e27_client.async_execute("system_get_trouble", timeout_s=_LIVE_TIMEOUT_S)
    if not result.ok:
        if extract_error_code(result.error) == E27ErrorCode.ELKERR_NOT_READY:
            pytest.skip("Panel reported not-ready for system_get_trouble.")
        pytest.fail(f"system_get_trouble failed: {describe_error(result.error)}")
    assert_payload_shape("system_get_trouble", result.data)

    error_code = _extract_error_code(result.data)
    if error_code == E27ErrorCode.ELKERR_NOT_READY:
        pytest.skip("Panel reported not-ready for system_get_trouble.")

    status = live_e27_client.state.system_status
    assert isinstance(status, dict)
    if not status:
        pytest.skip("system_get_trouble response was empty.")



from __future__ import annotations

import asyncio
import logging

import pytest

from elke27_lib.outbound import OutboundItem, OutboundPriority, OutboundQueue


def test_enqueue_stop_non_running_loop_and_debug() -> None:
    loop = asyncio.new_event_loop()
    try:
        logger = logging.getLogger("test_outbound")
        logger.setLevel(logging.DEBUG)
        sent: list[BaseException] = []

        def _on_fail(exc: BaseException) -> None:
            sent.append(exc)

        queue = OutboundQueue(loop=loop, send_fn=lambda _b: None, logger=logger)
        item = OutboundItem(
            payload=b"x",
            seq=1,
            kind="k",
            priority=OutboundPriority.HIGH,
            enqueued_at=0.0,
            on_fail=_on_fail,
        )
        queue.enqueue(item)
        assert queue.is_idle() is False
        queue.stop(fail_exc=RuntimeError("boom"))
        assert sent
    finally:
        loop.close()


@pytest.mark.asyncio
async def test_wait_idle_timeout_and_throttle_and_next_item() -> None:
    loop = asyncio.get_running_loop()
    queue = OutboundQueue(loop=loop, send_fn=lambda _b: None, min_interval_s=0)
    queue._high_q.put_nowait(
        OutboundItem(
            payload=b"x",
            seq=1,
            kind="k",
            priority=OutboundPriority.HIGH,
            enqueued_at=0.0,
        )
    )
    assert await queue.wait_idle(timeout_s=0.0) is False
    await queue._throttle()

    queue._stop_event.set()
    assert await queue._next_item() is None


@pytest.mark.asyncio
async def test_next_item_high_after_timeout(monkeypatch: pytest.MonkeyPatch) -> None:
    loop = asyncio.get_running_loop()
    queue = OutboundQueue(loop=loop, send_fn=lambda _b: None)

    async def _wait_for(coro, *args, **kwargs):  # type: ignore[no-untyped-def]
        coro.close()
        queue._high_q.put_nowait(
            OutboundItem(
                payload=b"x",
                seq=1,
                kind="k",
                priority=OutboundPriority.HIGH,
                enqueued_at=0.0,
            )
        )
        raise TimeoutError

    monkeypatch.setattr(asyncio, "wait_for", _wait_for)
    item = await queue._next_item()
    assert item is not None


@pytest.mark.asyncio
async def test_next_item_timeout_no_high(monkeypatch: pytest.MonkeyPatch) -> None:
    loop = asyncio.get_running_loop()
    queue = OutboundQueue(loop=loop, send_fn=lambda _b: None)

    async def _wait_for(coro, *args, **kwargs):  # type: ignore[no-untyped-def]
        coro.close()
        raise TimeoutError

    monkeypatch.setattr(asyncio, "wait_for", _wait_for)
    item = await queue._next_item()
    assert item is None


@pytest.mark.asyncio
async def test_run_handles_none_item() -> None:
    loop = asyncio.get_running_loop()
    queue = OutboundQueue(loop=loop, send_fn=lambda _b: None)

    async def _next_item():  # type: ignore[no-untyped-def]
        queue._stop_event.set()
        return None

    queue._next_item = _next_item  # type: ignore[assignment]
    await queue._run()


@pytest.mark.asyncio
async def test_run_send_success_and_failure() -> None:
    loop = asyncio.get_running_loop()
    sent: list[str] = []
    failed: list[str] = []

    def _send_ok(_b: bytes) -> None:
        sent.append("ok")

    logger = logging.getLogger("test_outbound_run")
    logger.setLevel(logging.DEBUG)
    queue = OutboundQueue(loop=loop, send_fn=_send_ok, logger=logger)
    queue.start()
    queue._normal_q.put_nowait(
        OutboundItem(
            payload=b"x",
            seq=1,
            kind="k",
            priority=OutboundPriority.NORMAL,
            enqueued_at=0.0,
            on_sent=lambda _t: sent.append("sent"),
        )
    )
    await queue.wait_idle(timeout_s=0.2)
    queue.stop()
    assert sent

    def _send_fail(_b: bytes) -> None:
        raise RuntimeError("boom")

    queue = OutboundQueue(loop=loop, send_fn=_send_fail)
    queue.start()
    queue._normal_q.put_nowait(
        OutboundItem(
            payload=b"x",
            seq=2,
            kind="k",
            priority=OutboundPriority.NORMAL,
            enqueued_at=0.0,
            on_fail=lambda _e: failed.append("fail"),
        )
    )
    await queue.wait_idle(timeout_s=0.2)
    queue.stop()
    assert failed


def test_drain_with_failure_queue_empty() -> None:
    loop = asyncio.new_event_loop()
    try:
        queue = OutboundQueue(loop=loop, send_fn=lambda _b: None)
        queue._drain_with_failure(None)

        queue._normal_q.put_nowait(
            OutboundItem(
                payload=b"x",
                seq=1,
                kind="k",
                priority=OutboundPriority.NORMAL,
                enqueued_at=0.0,
                on_fail=lambda _e: (_ for _ in ()).throw(RuntimeError("fail")),
            )
        )

        class _Queue:
            def empty(self):  # type: ignore[no-untyped-def]
                return False

            def get_nowait(self):  # type: ignore[no-untyped-def]
                raise asyncio.QueueEmpty

        queue._high_q = _Queue()  # type: ignore[assignment]
        queue._drain_with_failure(RuntimeError("boom"))
    finally:
        loop.close()



"""
elke27_lib/features/keypad.py

Feature module: keypad
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from elke27_lib.kernel import E27Kernel

from elke27_lib.handlers.keypad import (
    make_keypad_get_attribs_handler,
    make_keypad_get_configured_handler,
    make_keypad_get_table_info_handler,
)

ROUTE_KEYPAD_GET_CONFIGURED = ("keypad", "get_configured")
ROUTE_KEYPAD_GET_ATTRIBS = ("keypad", "get_attribs")
ROUTE_KEYPAD_GET_TABLE_INFO = ("keypad", "get_table_info")


def register(elk: E27Kernel) -> None:
    elk.register_handler(
        ROUTE_KEYPAD_GET_CONFIGURED,
        make_keypad_get_configured_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_KEYPAD_GET_ATTRIBS,
        make_keypad_get_attribs_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_KEYPAD_GET_TABLE_INFO,
        make_keypad_get_table_info_handler(elk.state, elk.emit, elk.now),
    )

    elk.register_request(
        ROUTE_KEYPAD_GET_CONFIGURED,
        build_keypad_get_configured_payload,
    )
    elk.register_request(
        ROUTE_KEYPAD_GET_ATTRIBS,
        build_keypad_get_attribs_payload,
    )
    elk.register_request(
        ROUTE_KEYPAD_GET_TABLE_INFO,
        build_keypad_get_table_info_payload,
    )


def build_keypad_get_configured_payload(*, block_id: int = 1, **_kwargs: Any) -> dict[str, object]:
    if block_id < 1:
        raise ValueError(
            f"build_keypad_get_configured_payload: block_id must be int >= 1 (got {block_id!r})"
        )
    return {"block_id": block_id}


def build_keypad_get_attribs_payload(*, keypad_id: int, **_kwargs: Any) -> dict[str, object]:
    if keypad_id < 1:
        raise ValueError(
            f"build_keypad_get_attribs_payload: keypad_id must be int >= 1 (got {keypad_id!r})"
        )
    return {"keypad_id": keypad_id}


def build_keypad_get_table_info_payload(**_kwargs: Any) -> dict[str, object]:
    return {}



"""Configured inventory filtering tests."""

from __future__ import annotations

from collections.abc import Callable, Mapping

from elke27_lib.client import Elke27Client
from elke27_lib.const import E27ErrorCode
from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    AreaConfiguredInventoryReady,
    AuthorizationRequiredEvent,
    Event,
    KeypadConfiguredInventoryReady,
    OutputConfiguredInventoryReady,
    UserConfiguredInventoryReady,
    ZoneConfiguredInventoryReady,
)
from elke27_lib.handlers import area as area_handler
from elke27_lib.handlers import zone as zone_handler
from elke27_lib.handlers.area import (
    make_area_get_attribs_handler,
    make_area_get_configured_handler,
)
from elke27_lib.handlers.keypad import make_keypad_get_configured_handler
from elke27_lib.handlers.user import make_user_get_configured_handler
from elke27_lib.handlers.zone import (
    make_zone_get_attribs_handler,
    make_zone_get_configured_handler,
)
from elke27_lib.states import PanelState
from test.helpers.dispatch import make_ctx
from test.helpers.internal import get_kernel, get_private


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[Event] = []

    def __call__(self, evt: Event, _ctx: DispatchContext) -> None:
        self.events.append(evt)


_Ctx = make_ctx


def test_zone_configured_paging_and_completion() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_zone_get_configured_handler(state, emit, now=lambda: 123.0)

    msg2 = {"zone": {"get_configured": {"block_count": 2, "zones": [1, 2, 28]}}}
    assert handler(msg2, _Ctx()) is True
    assert state.inventory.configured_zones == {1, 2, 28}
    assert state.inventory.configured_zones_complete is True
    assert any(isinstance(e, ZoneConfiguredInventoryReady) for e in emit.events)


def test_area_configured_paging_and_completion() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_area_get_configured_handler(state, emit, now=lambda: 123.0)

    msg2 = {"area": {"get_configured": {"block_id": 1, "block_count": 2, "areas": [1, 6]}}}
    assert handler(msg2, _Ctx()) is True
    assert state.inventory.configured_areas == {1, 6}
    assert state.inventory.configured_areas_complete is True
    assert any(isinstance(e, AreaConfiguredInventoryReady) for e in emit.events)


def test_user_configured_paging_and_completion() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_user_get_configured_handler(state, emit, now=lambda: 123.0)

    msg = {"user": {"get_configured": {"block_id": 1, "block_count": 1, "users": [1, 2]}}}
    assert handler(msg, _Ctx()) is True
    assert state.inventory.configured_users == {1, 2}
    assert state.inventory.configured_users_complete is True
    assert any(isinstance(e, UserConfiguredInventoryReady) for e in emit.events)


def test_keypad_configured_paging_and_completion() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_keypad_get_configured_handler(state, emit, now=lambda: 123.0)

    msg = {"keypad": {"get_configured": {"block_id": 1, "block_count": 1, "keypads": [1]}}}
    assert handler(msg, _Ctx()) is True
    assert state.inventory.configured_keypads == {1}
    assert state.inventory.configured_keypads_complete is True
    assert any(isinstance(e, KeypadConfiguredInventoryReady) for e in emit.events)


def test_zone_configured_auth_required_emits_event() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_zone_get_configured_handler(state, emit, now=lambda: 123.0)

    msg = {"zone": {"get_configured": {"error_code": E27ErrorCode.ELKERR_NOAUTH}}}
    assert handler(msg, _Ctx()) is True
    assert any(isinstance(e, AuthorizationRequiredEvent) for e in emit.events)
    assert state.inventory.configured_zones == set()


def test_client_filters_to_configured_inventory() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    kernel.state.inventory.configured_areas = {1}
    kernel.state.inventory.configured_zones = {2}
    kernel.state.get_or_create_area(1)
    kernel.state.get_or_create_area(2)
    kernel.state.get_or_create_zone(2)
    kernel.state.get_or_create_zone(3)

    assert list(client.areas) == [1]
    assert list(client.zones) == [2]


def test_area_get_attribs_name_normalization() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_area_get_attribs_handler(state, emit, now=lambda: 123.0)

    msg = {"area": {"get_attribs": {"area_id": 1, "name": "  Main  "}}}
    assert handler(msg, _Ctx()) is True
    assert state.areas[1].name == "Main"

    msg = {"area": {"get_attribs": {"area_id": 1, "name": "   "}}}
    assert handler(msg, _Ctx()) is True
    assert state.areas[1].name is None


def test_zone_get_attribs_name_normalization() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_zone_get_attribs_handler(state, emit, now=lambda: 123.0)

    msg = {"zone": {"get_attribs": {"zone_id": 1, "name": "\tFront  "}}}
    assert handler(msg, _Ctx()) is True
    assert state.zones[1].name == "Front"

    msg = {"zone": {"get_attribs": {"zone_id": 1, "name": ""}}}
    assert handler(msg, _Ctx()) is True
    assert state.zones[1].name is None


def test_attribs_request_queue_respects_limit() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    sent: list[tuple[tuple[str, str], dict[str, object]]] = []

    def _request(route: tuple[str, str], **kwargs: object):
        sent.append((route, dict(kwargs)))
        return 1

    kernel.request = _request

    inv = kernel.state.inventory
    inv.configured_areas = {1, 2, 3}
    inv.configured_zones = {1, 2}
    inv.configured_outputs = {1, 2}
    inv.configured_users = {1}
    inv.configured_keypads = {1}

    queue_bootstrap = get_private(client, "_queue_bootstrap_attribs")
    queue_bootstrap("area")
    queue_bootstrap("zone")
    queue_bootstrap("output")
    queue_bootstrap("user")
    queue_bootstrap("keypad")

    assert sent == [
        (("area", "get_attribs"), {"area_id": 1}),
        (("area", "get_attribs"), {"area_id": 2}),
        (("area", "get_attribs"), {"area_id": 3}),
        (("zone", "get_attribs"), {"zone_id": 1}),
        (("zone", "get_attribs"), {"zone_id": 2}),
        (("output", "get_attribs"), {"output_id": 1}),
        (("output", "get_attribs"), {"output_id": 2}),
        (("user", "get_attribs"), {"user_id": 1}),
        (("keypad", "get_attribs"), {"keypad_id": 1}),
    ]


def test_client_queues_output_attribs_on_inventory_ready() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    inv = kernel.state.inventory
    inv.configured_outputs = {1, 2}

    sent: list[tuple[tuple[str, str], dict[str, object]]] = []

    def _request(route: tuple[str, str], **kwargs: object):
        sent.append((route, dict(kwargs)))
        return 1

    kernel.request = _request

    handle_event = get_private(client, "_handle_kernel_event")
    handle_event(
        OutputConfiguredInventoryReady(
            kind=OutputConfiguredInventoryReady.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
        )
    )
    assert sent == [
        (("output", "get_attribs"), {"output_id": 1}),
        (("output", "get_attribs"), {"output_id": 2}),
        (("output", "get_status"), {"output_id": 1}),
        (("output", "get_status"), {"output_id": 2}),
    ]


def test_client_queues_user_keypad_attribs_on_inventory_ready() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    inv = kernel.state.inventory
    inv.configured_users = {1}
    inv.configured_keypads = {2}

    sent: list[tuple[tuple[str, str], dict[str, object]]] = []

    def _request(route: tuple[str, str], **kwargs: object):
        sent.append((route, dict(kwargs)))
        return 1

    kernel.request = _request

    handle_event = get_private(client, "_handle_kernel_event")
    handle_event(
        UserConfiguredInventoryReady(
            kind=UserConfiguredInventoryReady.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
        )
    )
    handle_event(
        KeypadConfiguredInventoryReady(
            kind=KeypadConfiguredInventoryReady.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
        )
    )
    assert sent == [
        (("user", "get_attribs"), {"user_id": 1}),
        (("keypad", "get_attribs"), {"keypad_id": 2}),
    ]


def test_area_get_attribs_auth_required_event() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_area_get_attribs_handler(state, emit, now=lambda: 123.0)

    msg = {"area": {"get_attribs": {"area_id": 1, "error_code": E27ErrorCode.ELKERR_NOAUTH}}}
    assert handler(msg, _Ctx()) is True
    assert any(isinstance(e, AuthorizationRequiredEvent) for e in emit.events)


def test_zone_get_attribs_auth_required_event() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_zone_get_attribs_handler(state, emit, now=lambda: 123.0)

    msg = {"zone": {"get_attribs": {"zone_id": 1, "error_code": E27ErrorCode.ELKERR_NOAUTH}}}
    assert handler(msg, _Ctx()) is True
    assert any(isinstance(e, AuthorizationRequiredEvent) for e in emit.events)


def test_zone_configured_ignores_bitmask_when_paged() -> None:
    payload = {
        "block_id": "1",
        "block_count": "2",
        "zones": [1, 2],
        "zone_mask": 0xFFFFFFFF,
    }
    warnings: list[str] = []
    extract_zone_ids: Callable[[Mapping[str, object], list[str]], list[int]] = get_private(
        zone_handler, "_extract_configured_zone_ids"
    )
    ids = extract_zone_ids(payload, warnings)
    assert ids == [1, 2]
    assert any("bitmask ignored" in w for w in warnings)


def test_area_configured_ignores_bitmask_when_paged() -> None:
    payload = {
        "block_id": "1",
        "block_count": "2",
        "areas": [1, 2],
        "area_mask": 0xFFFFFFFF,
    }
    warnings: list[str] = []
    extract_area_ids: Callable[[Mapping[str, object], list[str]], list[int]] = get_private(
        area_handler, "_extract_configured_area_ids"
    )
    ids = extract_area_ids(payload, warnings)
    assert ids == [1, 2]
    assert any("bitmask ignored" in w for w in warnings)



from __future__ import annotations

import asyncio
from collections.abc import Callable, Mapping
from typing import Any, cast

import pytest
from _pytest.monkeypatch import MonkeyPatch

from elke27_lib.client import Elke27Client
from elke27_lib.errors import NotAuthenticatedError
from elke27_lib.generators.registry import COMMANDS, CommandSpec
from elke27_lib.permissions import PermissionLevel
from test.helpers.internal import get_kernel, get_private


class _FakeSession:
    def __init__(self) -> None:
        self.sent: list[dict[str, Any]] = []

    def send_json(
        self,
        msg: dict[str, Any],
        *,
        priority: object = None,
        on_sent: Callable[[float], None] | None = None,
        on_fail: Callable[[BaseException], None] | None = None,
    ) -> None:
        del priority, on_fail
        self.sent.append(msg)
        if on_sent is not None:
            on_sent(0.0)


def _set_session(kernel: object, session: _FakeSession) -> None:
    cast(Any, kernel)._session = session


@pytest.mark.asyncio
async def test_async_execute_requires_session_for_area_set_status() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    fake_session = _FakeSession()
    _set_session(kernel, fake_session)

    result = await client.async_execute("area_set_status", area_id=1, chime=True)

    assert result.ok is False
    assert isinstance(result.error, NotAuthenticatedError)
    assert "area_set_status" in str(result.error)


@pytest.mark.asyncio
async def test_async_execute_does_not_block_master_command(monkeypatch: MonkeyPatch) -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    fake_session = _FakeSession()
    _set_session(kernel, fake_session)
    kernel.state.panel.session_id = 1
    area = kernel.state.get_or_create_area(1)
    area.arm_state = "disarmed"

    command_key = "flag_set_attribs"

    def _gen_flag_set_attribs(**kwargs: object) -> tuple[dict[str, object], tuple[str, str]]:
        return {"flag_id": kwargs["flag_id"]}, ("flag", "set_attribs")

    def _handler_flag_set_attribs(_msg: Mapping[str, object], _ctx: object) -> bool:
        return True

    spec = CommandSpec(
        key=command_key,
        domain="flag",
        command="set_attribs",
        generator=_gen_flag_set_attribs,
        handler=_handler_flag_set_attribs,
        min_permission=PermissionLevel.PLT_MASTER_USER_DISARMED,
        response_mode="single",
    )
    spec.generator.__name__ = "generator_flag_set_attribs"
    monkeypatch.setitem(COMMANDS, command_key, spec)

    task = asyncio.create_task(client.async_execute(command_key, flag_id=1, pin="1234"))
    await asyncio.sleep(0)

    sent = fake_session.sent[0]
    seq = sent["seq"]
    on_message = get_private(kernel, "_on_message")
    on_message({"seq": seq, "flag": {"set_attribs": {"error_code": 0}}})

    result = await task
    assert result.ok is True



from __future__ import annotations

import pytest

from elke27_lib.client import Elke27Client
from test.helpers.payload_validation import assert_payload_shape


async def _resolve_plug_dimmer_light_id(client: Elke27Client) -> int:
    expected_name = "plug dimmer"
    target_zw_id = 2

    configured = await client.async_execute("light_get_configured")
    if not configured.ok:
        pytest.fail(f"light_get_configured failed: {configured.error}")
    assert_payload_shape("light_get_configured", configured.data)

    lights = configured.data.get("lights") if configured.data else None
    if not isinstance(lights, list) or not lights:
        pytest.fail(f"No configured lights found; configured lights={lights!r}")

    for light_id in lights:
        if not isinstance(light_id, int) or light_id < 1:
            continue
        attribs = await client.async_execute("light_get_attribs", light_id=light_id)
        if not attribs.ok:
            continue
        assert_payload_shape("light_get_attribs", attribs.data)
        name = str((attribs.data or {}).get("name", "")).strip().lower()
        source_id = (attribs.data or {}).get("source_id")
        zw_id = source_id - 256 if isinstance(source_id, int) and source_id >= 256 else None
        if name == expected_name or zw_id == target_zw_id:
            return light_id

    pytest.fail(
        f"Unable to resolve '{expected_name}' / ZW-ID {target_zw_id} from configured lights={lights!r}"
    )


@pytest.mark.integration
@pytest.mark.asyncio
async def test_live_light_set_on_level_plug_dimmer_zw_id_2(
    live_e27_client: Elke27Client,
) -> None:
    target_id = await _resolve_plug_dimmer_light_id(live_e27_client)

    set_on = await live_e27_client.async_execute(
        "light_set_status",
        light_id=target_id,
        status="ON",
        level=50,
    )
    if not set_on.ok:
        pytest.fail(f"light_set_status ON failed: {set_on.error}")
    assert_payload_shape("light_set_status", set_on.data)

    status_on = await live_e27_client.async_execute("light_get_status", light_id=target_id)
    if not status_on.ok:
        pytest.fail(f"light_get_status after ON failed: {status_on.error}")
    assert_payload_shape("light_get_status", status_on.data)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_live_light_set_off_plug_dimmer_zw_id_2(live_e27_client: Elke27Client) -> None:
    target_id = await _resolve_plug_dimmer_light_id(live_e27_client)

    set_off = await live_e27_client.async_execute(
        "light_set_status",
        light_id=target_id,
        status="OFF",
        level=0,
    )
    if not set_off.ok:
        pytest.fail(f"light_set_status OFF failed: {set_off.error}")
    assert_payload_shape("light_set_status", set_off.data)

    status_off = await live_e27_client.async_execute("light_get_status", light_id=target_id)
    if not status_off.ok:
        pytest.fail(f"light_get_status after OFF failed: {status_off.error}")
    assert_payload_shape("light_get_status", status_off.data)



from __future__ import annotations

from collections.abc import Callable, Mapping

import elke27_lib.client as client_mod
from elke27_lib.const import E27ErrorCode
from elke27_lib.dispatcher import DispatchContext, PagedBlock
from elke27_lib.events import Event
from elke27_lib.handlers.keypad import (
    make_keypad_get_attribs_handler,
    make_keypad_get_configured_handler,
)
from elke27_lib.states import PanelState
from test.helpers.dispatch import make_ctx
from test.helpers.internal import get_private


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[Event] = []

    def __call__(self, evt: Event, _ctx: DispatchContext) -> None:
        self.events.append(evt)


_Ctx = make_ctx


def test_keypad_get_configured_merges_blocks() -> None:
    blocks = [
        PagedBlock(1, {"keypads": [1, 2]}),
        PagedBlock(2, {"keypads": [3]}),
    ]
    merge_fn: Callable[[list[PagedBlock], int], Mapping[str, object]] = get_private(
        client_mod, "_merge_configured_keypads"
    )
    merged = merge_fn(blocks, 2)
    assert merged == {"keypads": [1, 2, 3], "block_count": 2}


def test_keypad_handlers_store_snapshot() -> None:
    state = PanelState()
    emit = _EmitSpy()
    cfg_handler = make_keypad_get_configured_handler(state, emit, now=lambda: 123.0)
    attr_handler = make_keypad_get_attribs_handler(state, emit, now=lambda: 123.0)

    msg = {
        "keypad": {
            "get_configured": {
                "block_id": 1,
                "block_count": 1,
                "keypads": [1],
                "error_code": E27ErrorCode.ELKERR_NONE,
            }
        }
    }
    assert cfg_handler(msg, _Ctx()) is True
    assert state.inventory.configured_keypads == {1}

    msg = {
        "keypad": {
            "get_attribs": {
                "keypad_id": 1,
                "name": "Main Keypad",
                "area": 1,
                "error_code": E27ErrorCode.ELKERR_NONE,
            }
        }
    }
    assert attr_handler(msg, _Ctx()) is True
    assert state.keypads[1].name == "Main Keypad"
    assert state.keypads[1].area == 1



from __future__ import annotations

import asyncio
import logging
import os

import pytest

from elke27_lib.client import Elke27Client
from test.helpers.payload_validation import assert_payload_shape

LOG = logging.getLogger(__name__)


def _area_id() -> int:
    value = os.environ.get("ELKE27_AREA_ID") or "1"
    try:
        parsed = int(value)
    except ValueError:
        return 1
    return parsed if parsed > 0 else 1


def _require_interactive() -> None:
    if os.environ.get("ELKE27_INTERACTIVE", "").strip().lower() != "true":
        pytest.skip("Set ELKE27_INTERACTIVE=true to run chime tests.")


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_area_chime_on_ack(live_e27_client: Elke27Client) -> None:
    _require_interactive()
    area_id = _area_id()
    result = await live_e27_client.async_execute("area_set_status", area_id=area_id, chime=True)
    assert result.ok is True
    assert_payload_shape("area_set_status", result.data)


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_area_chime_off_ack(live_e27_client: Elke27Client) -> None:
    _require_interactive()
    area_id = _area_id()
    result = await live_e27_client.async_execute("area_set_status", area_id=area_id, chime=False)
    assert result.ok is True
    assert_payload_shape("area_set_status", result.data)


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_area_chime_on_persist(live_e27_client: Elke27Client) -> None:
    _require_interactive()
    if os.environ.get("ELKE27_HUMAN_TEST") != "1":
        pytest.skip("ELKE27_HUMAN_TEST not set; skipping persistent chime test.")
    area_id = _area_id()
    result = await live_e27_client.async_execute("area_set_status", area_id=area_id, chime=True)
    assert result.ok is True
    assert_payload_shape("area_set_status", result.data)


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_area_chime_state_roundtrip(live_e27_client: Elke27Client) -> None:
    _require_interactive()
    area_id = _area_id()
    area = live_e27_client.areas.get(area_id)
    snapshot_available = False
    original: bool | None = None
    if area is not None and area.chime is not None:
        snapshot_available = True
        original = bool(area.chime)
    if not snapshot_available:
        configured_result = await live_e27_client.async_execute("area_get_configured")
        if configured_result.ok:
            assert_payload_shape("area_get_configured", configured_result.data)
        await asyncio.sleep(0.5)
        area = live_e27_client.areas.get(area_id)
        if area is not None and area.chime is not None:
            snapshot_available = True
            original = bool(area.chime)
    if original is None:
        result = await live_e27_client.async_execute("area_get_status", area_id=area_id)
        if result.ok:
            assert_payload_shape("area_get_status", result.data)
        if result.ok and isinstance(result.data, dict) and "Chime" in result.data:
            original = bool(result.data["Chime"])
    if original is None:
        pytest.skip("Chime state not available in snapshot or response; skipping state roundtrip.")

    target = not original

    async def _wait_for_chime(expected: bool, timeout_s: float = 5.0) -> bool:
        if not snapshot_available:
            await asyncio.sleep(timeout_s)
            return False
        end = asyncio.get_running_loop().time() + timeout_s
        while asyncio.get_running_loop().time() < end:
            current_area = live_e27_client.areas.get(area_id)
            if current_area is not None and current_area.chime is expected:
                return True
            await asyncio.sleep(0.2)
        return False

    try:
        result = await live_e27_client.async_execute(
            "area_set_status", area_id=area_id, chime=target
        )
        if result.ok:
            assert_payload_shape("area_set_status", result.data)
        changed = await _wait_for_chime(target)
        if not changed and snapshot_available:
            LOG.warning("Chime state did not converge to %s within timeout.", target)
    finally:
        result = await live_e27_client.async_execute(
            "area_set_status", area_id=area_id, chime=original
        )
        if result.ok:
            assert_payload_shape("area_set_status", result.data)
        reverted = await _wait_for_chime(original)
        if not reverted and snapshot_available:
            LOG.warning("Chime state did not revert to %s within timeout.", original)



from __future__ import annotations

from typing import Any, cast

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import (
    ApiError,
    AuthorizationRequiredEvent,
    CsmSnapshotUpdated,
    TableCsmChanged,
)
from elke27_lib.handlers import log as log_handler
from elke27_lib.states import PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, _ctx: DispatchContext) -> None:
        self.events.append(evt)


def _any_event(spy: _EmitSpy, kind: type) -> bool:
    return any(isinstance(evt, kind) for evt in spy.events)


def test_log_handlers_store_status_and_errors() -> None:
    state = PanelState()
    emit = _EmitSpy()

    trouble = log_handler.make_log_get_trouble_handler(state, emit, now=lambda: 1.0)
    index = log_handler.make_log_get_index_handler(state, emit, now=lambda: 2.0)
    attribs = log_handler.make_log_get_attribs_handler(state, emit, now=lambda: 3.0)
    set_attribs = log_handler.make_log_set_attribs_handler(state, emit, now=lambda: 4.0)
    get_list = log_handler.make_log_get_list_handler(state, emit, now=lambda: 5.0)
    get_log = log_handler.make_log_get_log_handler(state, emit, now=lambda: 6.0)
    clear = log_handler.make_log_clear_handler(state, emit, now=lambda: 7.0)
    realloc = log_handler.make_log_realloc_handler(state, emit, now=lambda: 8.0)

    assert trouble({"nope": {}}, make_ctx()) is False
    assert trouble({"log": {}}, make_ctx()) is False
    assert index({"nope": {}}, make_ctx()) is False
    assert index({"log": {}}, make_ctx()) is False
    assert attribs({"nope": {}}, make_ctx()) is False
    assert attribs({"log": {}}, make_ctx()) is False
    assert set_attribs({"nope": {}}, make_ctx()) is False
    assert set_attribs({"log": {}}, make_ctx()) is False
    assert get_list({"nope": {}}, make_ctx()) is False
    assert get_list({"log": {}}, make_ctx()) is False
    assert get_log({"nope": {}}, make_ctx()) is False
    assert get_log({"log": {}}, make_ctx()) is False
    assert clear({"nope": {}}, make_ctx()) is False
    assert clear({"log": {}}, make_ctx()) is False
    assert realloc({"nope": {}}, make_ctx()) is False
    assert realloc({"log": {}}, make_ctx()) is False

    msg = {"log": {"get_trouble": {"error_code": 11008}}}
    assert trouble(msg, make_ctx(classification="BROADCAST")) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"get_index": {"error_code": 11008}}}
    assert index(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"get_attribs": {"error_code": 11008}}}
    assert attribs(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"set_attribs": {"error_code": 11008}}}
    assert set_attribs(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"get_list": {"error_code": 11008}}}
    assert get_list(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"get_log": {"error_code": 11008}}}
    assert get_log(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"clear": {"error_code": 11008}}}
    assert clear(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"realloc": {"error_code": 11008}}}
    assert realloc(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"get_trouble": {"error_code": 5}}}
    assert trouble(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"log": {"get_trouble": {"foo": "bar"}}}
    assert trouble(msg, make_ctx()) is True
    assert cast(dict[str, Any], state.log_status["get_trouble"])["foo"] == "bar"
    assert state.panel.last_message_at == 1.0

    msg = {"log": {"get_index": {"idx": 1}}}
    assert index(msg, make_ctx()) is True
    assert cast(dict[str, Any], state.log_status["get_index"])["idx"] == 1
    assert state.panel.last_message_at == 2.0

    msg = {"log": {"get_attribs": {"a": 1}}}
    assert attribs(msg, make_ctx()) is True
    assert cast(dict[str, Any], state.log_status["get_attribs"])["a"] == 1

    msg = {"log": {"set_attribs": {"b": 2}}}
    assert set_attribs(msg, make_ctx()) is True
    assert cast(dict[str, Any], state.log_status["set_attribs"])["b"] == 2

    msg = {"log": {"get_list": {"c": 3}}}
    assert get_list(msg, make_ctx()) is True
    assert cast(dict[str, Any], state.log_status["get_list"])["c"] == 3

    msg = {"log": {"get_log": {"d": 4}}}
    assert get_log(msg, make_ctx()) is True
    assert cast(dict[str, Any], state.log_status["get_log"])["d"] == 4

    msg = {"log": {"clear": {"e": 5}}}
    assert clear(msg, make_ctx()) is True
    assert cast(dict[str, Any], state.log_status["clear"])["e"] == 5

    msg = {"log": {"realloc": {"f": 6}}}
    assert realloc(msg, make_ctx()) is True
    assert cast(dict[str, Any], state.log_status["realloc"])["f"] == 6


def test_log_table_info_and_helpers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = log_handler.make_log_get_table_info_handler(state, emit, now=lambda: 9.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"log": {}}, make_ctx()) is False

    msg = {"log": {"get_table_info": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"log": {"get_table_info": {"table_csm": 12, "table_elements": 2}}}
    assert handler(msg, make_ctx()) is True
    assert state.table_info_by_domain["log"]["table_elements"] == 2
    assert _any_event(emit, TableCsmChanged)
    assert _any_event(emit, CsmSnapshotUpdated)

    assert log_handler._extract_table_csm({"table_csm": True}, domain="log") is None
    assert log_handler._extract_table_csm({"table_csm": 2}, domain="log") == 2
    assert log_handler._extract_table_csm({"table_csm": "3"}, domain="log") == 3
    assert log_handler._extract_table_csm({"table_csm": "bad"}, domain="log") is None
    assert log_handler._extract_table_csm({"other": 1}, domain="log") is None



from __future__ import annotations

import logging

import pytest

from elke27_lib.dispatcher import (
    DispatchContext,
    Dispatcher,
    MessageKind,
    PagedRouteSpec,
    PagedTransfer,
    PagedTransferKey,
    PendingRequest,
    _payload_preview,
)


def test_payload_preview_non_serializable_and_truncate() -> None:
    msg = {"a": {1, 2, 3}}
    text = _payload_preview(msg, limit=5)
    assert text.endswith("...")


def test_register_domain_and_unregister() -> None:
    dispatcher = Dispatcher()

    def handler(_msg, _ctx):  # type: ignore[no-untyped-def]
        return True

    def handler2(_msg, _ctx):  # type: ignore[no-untyped-def]
        return False

    dispatcher.register_domain("zone", handler)
    assert ("zone", "__root__") in dispatcher._handlers

    dispatcher.unregister(("zone", "__root__"), handler2)
    dispatcher.unregister(("zone", "__root__"), handler)
    assert ("zone", "__root__") not in dispatcher._handlers
    dispatcher.unregister(("zone", "__root__"), handler)


def test_pending_count_and_drop() -> None:
    dispatcher = Dispatcher()
    pending = PendingRequest(seq=1)
    dispatcher.add_pending(pending)
    assert dispatcher.pending_count() == 1
    dispatcher.drop_pending(1)
    assert dispatcher.pending_count() == 0


def test_extract_route_errors() -> None:
    dispatcher = Dispatcher()
    route, errors = dispatcher._extract_route({"seq": 1})
    assert errors and route[0] == "__root__"

    route, errors = dispatcher._extract_route({"a": {}, "b": {}})
    assert errors and route[1] == "__multi__"

    route, errors = dispatcher._extract_route({"zone": {}})
    assert errors and route[1] == "__empty__"

    route, errors = dispatcher._extract_route({"zone": {"a": 1, "b": 2}})
    assert errors and route[1] == "__root__"

    route, errors = dispatcher._extract_route({"zone": 1})
    assert errors and route[1] == "__value__"


def test_domain_root_error_and_root_error_envelope() -> None:
    dispatcher = Dispatcher()
    assert dispatcher._is_domain_root_error({"error_code": "abc"}) is False
    assert dispatcher._is_domain_root_error({"error_code": "5", "foo": {"bar": 1}}) is False
    assert dispatcher._is_domain_root_error({"error_code": "5"}) is True

    assert dispatcher._is_root_error_envelope({"error_code": "abc", "error_message": "x"}) is False
    assert dispatcher._is_root_error_envelope({"error_code": 5, "error_message": ""}) is False
    assert dispatcher._is_root_error_envelope({"error_code": 5, "error_message": "bad"}) is True


def test_classify_kind_variants() -> None:
    dispatcher = Dispatcher()
    seq, kind, errors = dispatcher._classify_kind({})
    assert kind is MessageKind.UNKNOWN and not errors
    seq, kind, errors = dispatcher._classify_kind({"seq": "x"})
    assert kind is MessageKind.UNKNOWN and errors
    seq, kind, errors = dispatcher._classify_kind({"seq": 0})
    assert kind is MessageKind.BROADCAST
    seq, kind, errors = dispatcher._classify_kind({"seq": 5})
    assert kind is MessageKind.DIRECTED
    seq, kind, errors = dispatcher._classify_kind({"seq": -1})
    assert kind is MessageKind.UNKNOWN and errors


def test_paged_transfer_key_and_extract_payload() -> None:
    dispatcher = Dispatcher()
    key = PagedTransferKey(session_id=1, transfer_id=2, route=("zone", "get_configured"))
    pending = PendingRequest(seq=2, opaque=key)
    assert dispatcher._paged_transfer_key(("zone", "get_configured"), 1, pending) is key

    pending2 = PendingRequest(seq=3)
    key2 = dispatcher._paged_transfer_key(("zone", "get_configured"), 1, pending2)
    assert key2 is not None

    assert dispatcher._paged_transfer_key(("zone", "get_configured"), None, None) is None

    assert dispatcher._extract_paged_payload({"zone": 1}, ("zone", "get_configured")) is None
    assert (
        dispatcher._extract_paged_payload(
            {"zone": {"get_configured": 1}}, ("zone", "get_configured")
        )
        is None
    )


def test_expire_paged_transfers_and_abort() -> None:
    dispatcher = Dispatcher(now=lambda: 10.0, paged_timeout_s=1.0)
    route = ("zone", "get_configured")
    dispatcher._paged_routes[route] = PagedRouteSpec(
        merge_fn=lambda *_a: {}, request_block=None, timeout_s=1.0
    )
    key = PagedTransferKey(session_id=1, transfer_id=1, route=route)
    dispatcher._paged_transfers[key] = PagedTransfer(
        key=key, total_count=1, created_at=0.0, last_update_at=0.0
    )
    dispatcher._expire_paged_transfers()
    assert key not in dispatcher._paged_transfers
    dispatcher._paged_transfers[key] = PagedTransfer(
        key=key, total_count=1, created_at=0.0, last_update_at=9.5
    )
    dispatcher.abort_paged_transfers()
    assert dispatcher._paged_transfers == {}


def test_maybe_reassemble_paged_branches(monkeypatch: pytest.MonkeyPatch) -> None:
    dispatcher = Dispatcher()
    route = ("zone", "get_configured")
    dispatcher.register_paged(route, merge_fn=lambda blocks, total: {"blocks": len(blocks)})

    ctx = DispatchContext(
        kind=MessageKind.DIRECTED,
        seq=1,
        session_id=1,
        route=route,
        classification="RESPONSE",
        response_match=None,
        raw_route=route,
    )
    response_match = PendingRequest(seq=1)
    msg = {"zone": {"get_configured": {"block_id": 0}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is None

    msg = {"zone": {"get_configured": {"block_id": 1, "block_count": 0}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is None

    msg = {"zone": {"get_configured": {"block_id": 3, "block_count": 2}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is None

    msg = {"zone": {"get_configured": 1}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is msg

    msg = {"zone": {"error_code": 11008, "get_configured": {"block_id": 1, "block_count": 1}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is msg

    msg = {"zone": {"get_configured": {"block_id": 1, "block_count": 1, "error_code": 11008}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is msg

    def _req(_block, _key):  # type: ignore[no-untyped-def]
        raise RuntimeError("fail")

    dispatcher._paged_routes[route] = PagedRouteSpec(
        merge_fn=lambda blocks, total: {"blocks": len(blocks)}, request_block=_req, timeout_s=1.0
    )
    msg = {"zone": {"get_configured": {"block_id": 1, "block_count": 2}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is None

    transfer_key = dispatcher._paged_transfer_key(route, 1, response_match)
    assert transfer_key is not None
    dispatcher._paged_transfers[transfer_key] = PagedTransfer(
        key=transfer_key, total_count=1, created_at=0.0, last_update_at=0.0
    )
    msg = {"zone": {"get_configured": {"block_id": 1, "block_count": 2}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is msg

    dispatcher._paged_transfers[transfer_key] = PagedTransfer(
        key=transfer_key, total_count=1, created_at=0.0, last_update_at=0.0, received_blocks={1: {}}
    )
    msg = {"zone": {"get_configured": {"block_id": 1, "block_count": 1}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is None

    def _req_ok(_block, _key):  # type: ignore[no-untyped-def]
        return None

    dispatcher._paged_routes[route] = PagedRouteSpec(
        merge_fn=lambda blocks, total: {"blocks": len(blocks)}, request_block=_req_ok, timeout_s=1.0
    )
    dispatcher._paged_transfers[transfer_key] = PagedTransfer(
        key=transfer_key,
        total_count=None,
        created_at=0.0,
        last_update_at=0.0,
        requested_blocks={1},
    )
    msg = {"zone": {"get_configured": {"block_id": 1, "block_count": 2}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is None


def test_dispatch_zone_debug_and_error_emit(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    dispatcher = Dispatcher()
    route = ("zone", "get_configured")
    dispatcher.register_paged(route, merge_fn=lambda blocks, total: {"blocks": len(blocks)})

    def _handler(_msg, _ctx):  # type: ignore[no-untyped-def]
        return True

    dispatcher.register(route, _handler)
    caplog.set_level(logging.DEBUG, logger="elke27_lib.dispatcher")
    msg = {"seq": "bad", "zone": {"get_configured": {"block_id": 0}}}
    result = dispatcher.dispatch(msg)
    assert result.handled is True


def test_dispatch_paged_emits_errors_when_partial(monkeypatch: pytest.MonkeyPatch) -> None:
    dispatcher = Dispatcher()
    route = ("zone", "__root__")
    dispatcher.register_paged(route, merge_fn=lambda blocks, total: {"blocks": len(blocks)})

    called: dict[str, int] = {"count": 0}

    def _emit_errors(_ctx, _errors, msg=None):  # type: ignore[no-untyped-def]
        called["count"] += 1

    monkeypatch.setattr(dispatcher, "_emit_errors", _emit_errors)

    dispatcher.add_pending(PendingRequest(seq=1))
    msg = {
        "seq": 1,
        "session_id": 1,
        "zone": {
            "__root__": {"block_id": 0, "block_count": 2},
            "other": {},
        },
    }
    result = dispatcher.dispatch(msg)

    assert result.handled is True
    assert called["count"] == 1


def test_maybe_reassemble_paged_skips_requested_block(monkeypatch: pytest.MonkeyPatch) -> None:
    dispatcher = Dispatcher()
    route = ("zone", "get_configured")

    def _req(_block, _key):  # type: ignore[no-untyped-def]
        return None

    dispatcher.register_paged(
        route, merge_fn=lambda blocks, total: {"blocks": len(blocks)}, request_block=_req
    )

    ctx = DispatchContext(
        kind=MessageKind.DIRECTED,
        seq=2,
        session_id=1,
        route=route,
        classification="RESPONSE",
        response_match=None,
        raw_route=route,
    )
    response_match = PendingRequest(seq=2)
    transfer_key = dispatcher._paged_transfer_key(route, 1, response_match)
    assert transfer_key is not None

    dispatcher._paged_transfers[transfer_key] = PagedTransfer(
        key=transfer_key,
        total_count=2,
        created_at=0.0,
        last_update_at=0.0,
        requested_blocks={1},
        received_blocks={},
    )

    msg = {"zone": {"get_configured": {"block_id": 2, "block_count": 2}}}
    assert dispatcher._maybe_reassemble_paged(msg, ctx, response_match) is None



from __future__ import annotations

import asyncio
import os

from elke27_lib import linking
from elke27_lib.client import Elke27Client

TARGET_NAME = "plug dimmer"
TARGET_ID_HINT = 2


def req(name: str) -> str:
    value = os.environ.get(name, "").strip()
    if not value:
        raise RuntimeError(f"Missing env var: {name}")
    return value


async def _main() -> None:
    host = req("ELKE27_HOST")
    port = int(os.environ.get("ELKE27_PORT", "2101"))
    access_code = req("ELKE27_ACCESS_CODE")
    passphrase = req("ELKE27_PASSPHRASE")
    mn = req("ELKE27_MN")
    sn = req("ELKE27_SN")
    fwver = req("ELKE27_FWVER")
    hwver = req("ELKE27_HWVER")
    osver = req("ELKE27_OSVER")

    identity = linking.E27Identity(mn=mn, sn=sn, fwver=fwver, hwver=hwver, osver=osver)
    client = Elke27Client()

    link_keys = await client.async_link(
        host,
        port,
        access_code=access_code,
        passphrase=passphrase,
        client_identity=identity,
        timeout_s=10.0,
    )
    await client.async_connect(host, port, link_keys)

    try:
        configured = await client.async_execute("light_get_configured", timeout_s=10.0)
        if not configured.ok or not configured.data:
            raise RuntimeError(f"light_get_configured failed: {configured.error}")

        ids = configured.data.get("lights")
        if not isinstance(ids, list):
            raise RuntimeError(f"Unexpected light_get_configured payload: {configured.data}")

        names: dict[int, str] = {}
        target_id: int | None = None
        for item in ids:
            if not isinstance(item, int) or item < 1:
                continue
            attribs = await client.async_execute("light_get_attribs", light_id=item, timeout_s=10.0)
            if not attribs.ok or not attribs.data:
                continue
            name = str(attribs.data.get("name") or "").strip()
            names[item] = name
            if name.lower() == TARGET_NAME:
                target_id = item

        if target_id is None and TARGET_ID_HINT in names:
            target_id = TARGET_ID_HINT
        if target_id is None:
            raise RuntimeError(f"Did not find '{TARGET_NAME}'. Lights seen: {names}")

        print(f"Using light_id={target_id} name='{names.get(target_id, '')}'")

        on = await client.async_execute(
            "light_set_status",
            light_id=target_id,
            status="ON",
            level=100,
            timeout_s=10.0,
        )
        if not on.ok:
            raise RuntimeError(f"light_set_status ON failed: {on.error}")
        print(f"ON ack: {on.data}")

        status_on = await client.async_execute(
            "light_get_status", light_id=target_id, timeout_s=10.0
        )
        print(f"After ON status: ok={status_on.ok} data={status_on.data} err={status_on.error}")

        off = await client.async_execute(
            "light_set_status",
            light_id=target_id,
            status="OFF",
            level=0,
            timeout_s=10.0,
        )
        if not off.ok:
            raise RuntimeError(f"light_set_status OFF failed: {off.error}")
        print(f"OFF ack: {off.data}")

        status_off = await client.async_execute(
            "light_get_status", light_id=target_id, timeout_s=10.0
        )
        print(f"After OFF status: ok={status_off.ok} data={status_off.data} err={status_off.error}")
    finally:
        await client.async_disconnect()


if __name__ == "__main__":
    asyncio.run(_main())



from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import Any, cast

import pytest

from elke27_lib.client import Elke27Client
from elke27_lib.errors import E27Timeout
from test.helpers.internal import get_kernel, get_private


class _FakeSession:
    cfg: object

    def __init__(self) -> None:
        self.cfg = type("_Cfg", (), {"host": "test-host", "port": 1})()
        self.sent: list[dict[str, Any]] = []

    def send_json(
        self,
        msg: dict[str, Any],
        *,
        priority: object = None,
        on_sent: Callable[[float], None] | None = None,
        on_fail: Callable[[BaseException], None] | None = None,
    ) -> None:
        del priority, on_fail
        self.sent.append(msg)
        if on_sent is not None:
            on_sent(0.0)


def _set_session(kernel: object, session: _FakeSession) -> None:
    cast(Any, kernel)._session = session


async def _wait_for_sent(session: _FakeSession, count: int, *, timeout_s: float = 0.1) -> None:
    loop = asyncio.get_running_loop()
    end = loop.time() + timeout_s
    while len(session.sent) < count and loop.time() < end:
        await asyncio.sleep(0)


@pytest.mark.asyncio
async def test_async_execute_paged_blocks_merges() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    fake_session = _FakeSession()
    _set_session(kernel, fake_session)
    kernel.state.panel.session_id = 1

    task = asyncio.create_task(client.async_execute("zone_get_configured"))
    await asyncio.sleep(0)

    sent1 = fake_session.sent[0]
    seq1 = sent1["seq"]
    assert sent1["zone"]["get_configured"]["block_id"] == 1

    on_message = get_private(kernel, "_on_message")
    on_message(
        {
            "seq": seq1,
            "zone": {"get_configured": {"block_id": 1, "block_count": 3, "zones": [1, 2]}},
        }
    )
    await _wait_for_sent(fake_session, 2)

    sent2 = fake_session.sent[1]
    seq2 = sent2["seq"]
    assert sent2["zone"]["get_configured"]["block_id"] == 2

    on_message = get_private(kernel, "_on_message")
    on_message(
        {
            "seq": seq2,
            "zone": {"get_configured": {"block_id": 2, "block_count": 3, "zones": [3]}},
        }
    )
    await _wait_for_sent(fake_session, 3)

    sent3 = fake_session.sent[2]
    seq3 = sent3["seq"]
    assert sent3["zone"]["get_configured"]["block_id"] == 3

    on_message = get_private(kernel, "_on_message")
    on_message(
        {
            "seq": seq3,
            "zone": {"get_configured": {"block_id": 3, "block_count": 3, "zones": [4, 5]}},
        }
    )

    result = await task
    assert result.ok is True
    assert result.data == {"zones": [1, 2, 3, 4, 5], "block_count": 3}
    pending = get_private(kernel, "_pending_responses")
    assert pending.pending_count() == 0


@pytest.mark.asyncio
async def test_async_execute_paged_timeout_on_block() -> None:
    client = Elke27Client()
    kernel = get_kernel(client)
    fake_session = _FakeSession()
    _set_session(kernel, fake_session)
    kernel.state.panel.session_id = 1

    task = asyncio.create_task(client.async_execute("zone_get_configured", timeout_s=0.01))
    await asyncio.sleep(0)

    sent1 = fake_session.sent[0]
    seq1 = sent1["seq"]

    on_message = get_private(kernel, "_on_message")
    on_message(
        {
            "seq": seq1,
            "zone": {"get_configured": {"block_id": 1, "block_count": 2, "zones": [1]}},
        }
    )

    result = await task
    assert result.ok is False
    assert isinstance(result.error, E27Timeout)
    pending = get_private(kernel, "_pending_responses")
    assert pending.pending_count() == 0



"""Redaction helpers for diagnostics-safe logging."""

from __future__ import annotations

from collections.abc import Iterable, Mapping
from dataclasses import asdict, is_dataclass
from enum import Enum
from typing import Any, cast

from .const import REDACT_DIAGNOSTICS

_SECRET_KEYS = {
    "passphrase",
    "access_code",
    "accesscode",
    "pin",
    "linkkey",
    "link_keys",
    "linkkey_hex",
    "linkhmac",
    "linkhmac_hex",
    "tempkey",
    "tempkey_hex",
    "session_key",
    "session_keys",
    "token",
    "hmac",
    "secret",
    "key",
    "keys",
}


def _should_redact(key: str) -> bool:
    key_l = key.lower()
    if key_l in _SECRET_KEYS:
        return True
    if "pass" in key_l or "token" in key_l or "secret" in key_l:
        return True
    return bool("key" in key_l and "monkey" not in key_l)


def _normalize_mapping(data: Mapping[str, Any]) -> dict[str, Any]:
    out: dict[str, Any] = {}
    for key, value in data.items():
        out[str(key)] = _normalize_for_diagnostics(value)
    return out


def _redact_mapping(data: Mapping[str, Any]) -> dict[str, Any]:
    out: dict[str, Any] = {}
    for key, value in data.items():
        key_str = str(key)
        if _should_redact(key_str):
            out[key_str] = "***"
        else:
            out[key_str] = redact_for_diagnostics(value)
    return out


def _normalize_for_diagnostics(obj: Any) -> Any:
    if obj is None:
        return None
    if is_dataclass(obj) and not isinstance(obj, type):
        return _normalize_mapping(asdict(obj))
    if isinstance(obj, Mapping):
        return _normalize_mapping(cast(Mapping[str, Any], obj))
    if isinstance(obj, (list, tuple, set)):
        items: list[object] = list(cast(Iterable[object], obj))
        return [_normalize_for_diagnostics(item) for item in items]
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, (bytes, bytearray)):
        return f"<bytes:{len(obj)}>"
    if isinstance(obj, (int, float, str, bool)):
        return obj
    return repr(obj)


def redact_for_diagnostics(obj: Any) -> Any:
    """
    Return a JSON-serializable structure with likely secrets redacted.

    This is intentionally conservative and may redact additional fields to
    ensure diagnostics are safe for logs.
    """

    if not REDACT_DIAGNOSTICS:
        return _normalize_for_diagnostics(obj)
    if obj is None:
        return None
    if is_dataclass(obj) and not isinstance(obj, type):
        return _redact_mapping(asdict(obj))
    if isinstance(obj, Mapping):
        return _redact_mapping(cast(Mapping[str, Any], obj))
    if isinstance(obj, (list, tuple, set)):
        items: list[object] = list(cast(Iterable[object], obj))
        return [redact_for_diagnostics(item) for item in items]
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, (bytes, bytearray)):
        return f"<bytes:{len(obj)}>"
    if isinstance(obj, (int, float, str, bool)):
        return obj
    return repr(obj)



import logging

from _pytest.logging import LogCaptureFixture

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import Event
from elke27_lib.handlers.area import make_area_get_configured_handler
from elke27_lib.states import PanelState
from test.helpers.dispatch import make_ctx


def test_area_get_configured_block_id_is_used(caplog: LogCaptureFixture) -> None:
    state = PanelState()
    emitted: list[Event] = []

    def _emit(evt: Event, _ctx: DispatchContext) -> None:
        emitted.append(evt)

    handler = make_area_get_configured_handler(state, _emit, lambda: 0.0)
    msg = {"area": {"get_configured": {"block_id": 1, "block_count": 1, "areas": [1, 2]}}}

    with caplog.at_level(logging.WARNING):
        assert handler(msg, make_ctx()) is True

    assert not caplog.records
    assert state.inventory.configured_areas == {1, 2}



"""
elke27_lib/events.py

Event dataclasses v0 (clean model).

Rules:
- Handlers construct event objects directly, with placeholder header fields.
- kernel.emit() stamps the authoritative header fields from DispatchContext and enqueues.
- No payload helper functions (cleanest possible API surface).
"""

from __future__ import annotations

from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from typing_extensions import override
else:

    def override(func):  # type: ignore[no-redef]
        return func


from elke27_lib.types import CsmSnapshot

RouteKey = tuple[str, str]


# -------------------------
# Common header (stamped by kernel.emit)
# -------------------------


@dataclass(frozen=True, slots=True)
class Event:
    # Common header fields (kernel.emit overwrites these unconditionally)
    kind: str
    at: float
    seq: int | None
    classification: str
    route: RouteKey
    session_id: int | None

    @property
    def domain(self) -> str:
        return self.route[0]


# Placeholder header values for handlers (optional convenience constants)
UNSET_ROUTE: RouteKey = ("__unset__", "__unset__")
UNSET_AT: float = 0.0
UNSET_SEQ: int | None = None
UNSET_CLASSIFICATION: str = "UNKNOWN"
UNSET_SESSION_ID: int | None = None


# -------------------------
# Connection lifecycle
# -------------------------


@dataclass(frozen=True, slots=True)
class ConnectionStateChanged(Event):
    KIND: ClassVar[str] = "connection_state_changed"

    connected: bool
    reason: str | None = None
    error_type: str | None = None


# -------------------------
# Area events
# -------------------------


@dataclass(frozen=True, slots=True)
class AreaStatusUpdated(Event):
    KIND: ClassVar[str] = "area_status_updated"

    area_id: int
    changed_fields: tuple[str, ...]  # sorted tuple for deterministic tests/logs


@dataclass(frozen=True, slots=True)
class AreaAttribsUpdated(Event):
    KIND: ClassVar[str] = "area_attribs_updated"

    area_id: int
    changed_fields: tuple[str, ...]


@dataclass(frozen=True, slots=True)
class AreaConfiguredUpdated(Event):
    KIND: ClassVar[str] = "area_configured_updated"

    configured_ids: tuple[int, ...]


@dataclass(frozen=True, slots=True)
class AreaConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "area_configured_inventory_ready"


# -------------------------
# Zone events
# -------------------------


@dataclass(frozen=True, slots=True)
class ZoneConfiguredUpdated(Event):
    KIND: ClassVar[str] = "zone_configured_updated"

    configured_ids: tuple[int, ...]


@dataclass(frozen=True, slots=True)
class ZoneConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "zone_configured_inventory_ready"


@dataclass(frozen=True, slots=True)
class ZoneStatusUpdated(Event):
    KIND: ClassVar[str] = "zone_status_updated"

    zone_id: int
    changed_fields: tuple[str, ...]


@dataclass(frozen=True, slots=True)
class ZonesStatusBulkUpdated(Event):
    KIND: ClassVar[str] = "zones_status_bulk_updated"

    updated_count: int
    updated_ids: tuple[int, ...]


# -------------------------
# Zone definitions
# -------------------------


@dataclass(frozen=True, slots=True)
class ZoneDefsUpdated(Event):
    KIND: ClassVar[str] = "zone_defs_updated"

    count: int
    updated_ids: tuple[int, ...]


@dataclass(frozen=True, slots=True)
class ZoneDefFlagsUpdated(Event):
    KIND: ClassVar[str] = "zone_def_flags_updated"

    count: int


# -------------------------
# Zone attribs
# -------------------------


@dataclass(frozen=True, slots=True)
class ZoneAttribsUpdated(Event):
    KIND: ClassVar[str] = "zone_attribs_updated"

    zone_id: int
    changed_fields: tuple[str, ...]


# -------------------------
# Area troubles
# -------------------------


@dataclass(frozen=True, slots=True)
class AreaTroublesUpdated(Event):
    KIND: ClassVar[str] = "area_troubles_updated"

    area_id: int | None
    troubles: tuple[str, ...]


# -------------------------
# Output events
# -------------------------


@dataclass(frozen=True, slots=True)
class OutputConfiguredUpdated(Event):
    KIND: ClassVar[str] = "output_configured_updated"

    configured_ids: tuple[int, ...]


@dataclass(frozen=True, slots=True)
class OutputConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "output_configured_inventory_ready"


@dataclass(frozen=True, slots=True)
class LightConfiguredUpdated(Event):
    KIND: ClassVar[str] = "light_configured_updated"

    configured_ids: tuple[int, ...]


@dataclass(frozen=True, slots=True)
class LightConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "light_configured_inventory_ready"


@dataclass(frozen=True, slots=True)
class BarrierConfiguredUpdated(Event):
    KIND: ClassVar[str] = "barrier_configured_updated"

    configured_ids: tuple[int, ...]


@dataclass(frozen=True, slots=True)
class BarrierConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "barrier_configured_inventory_ready"


@dataclass(frozen=True, slots=True)
class LockConfiguredUpdated(Event):
    KIND: ClassVar[str] = "lock_configured_updated"

    configured_ids: tuple[int, ...]


@dataclass(frozen=True, slots=True)
class LockConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "lock_configured_inventory_ready"


@dataclass(frozen=True, slots=True)
class TstatConfiguredUpdated(Event):
    KIND: ClassVar[str] = "tstat_configured_updated"

    configured_ids: tuple[int, ...]


@dataclass(frozen=True, slots=True)
class TstatConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "tstat_configured_inventory_ready"


@dataclass(frozen=True, slots=True)
class UserConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "user_configured_inventory_ready"


@dataclass(frozen=True, slots=True)
class KeypadConfiguredInventoryReady(Event):
    KIND: ClassVar[str] = "keypad_configured_inventory_ready"


@dataclass(frozen=True, slots=True)
class OutputStatusUpdated(Event):
    KIND: ClassVar[str] = "output_status_updated"

    output_id: int
    status: str | None
    on: bool | None


@dataclass(frozen=True, slots=True)
class OutputsStatusBulkUpdated(Event):
    KIND: ClassVar[str] = "outputs_status_bulk_updated"

    updated_count: int
    updated_ids: tuple[int, ...]


# -------------------------
# Tstat events
# -------------------------


@dataclass(frozen=True, slots=True)
class TstatStatusUpdated(Event):
    KIND: ClassVar[str] = "tstat_status_updated"

    tstat_id: int
    mode: str | None
    fan_mode: str | None
    temperature: int | None


@dataclass(frozen=True, slots=True)
class LightStatusUpdated(Event):
    KIND: ClassVar[str] = "light_status_updated"

    light_id: int
    status: str | None
    on: bool | None
    level: int | None


@dataclass(frozen=True, slots=True)
class BarrierStatusUpdated(Event):
    KIND: ClassVar[str] = "barrier_status_updated"

    barrier_id: int
    status: str | None


@dataclass(frozen=True, slots=True)
class LockStatusUpdated(Event):
    KIND: ClassVar[str] = "lock_status_updated"

    lock_id: int
    status: str | None
    locked: bool | None


# -------------------------
# Network events
# -------------------------


@dataclass(frozen=True, slots=True)
class NetworkSsidResultsUpdated(Event):
    KIND: ClassVar[str] = "network_ssid_results_updated"

    count: int
    ssids: tuple[str, ...]


@dataclass(frozen=True, slots=True)
class NetworkRssiUpdated(Event):
    KIND: ClassVar[str] = "network_rssi_updated"

    rssi: int | None


# -------------------------
# Table info events
# -------------------------


@dataclass(frozen=True, slots=True)
class AreaTableInfoUpdated(Event):
    KIND: ClassVar[str] = "area_table_info_updated"

    table_elements: int | None
    increment_size: int | None
    table_csm: int | None


@dataclass(frozen=True, slots=True)
class ZoneTableInfoUpdated(Event):
    KIND: ClassVar[str] = "zone_table_info_updated"

    table_elements: int | None
    increment_size: int | None
    table_csm: int | None


@dataclass(frozen=True, slots=True)
class OutputTableInfoUpdated(Event):
    KIND: ClassVar[str] = "output_table_info_updated"

    table_elements: int | None
    increment_size: int | None
    table_csm: int | None


@dataclass(frozen=True, slots=True)
class TstatTableInfoUpdated(Event):
    KIND: ClassVar[str] = "tstat_table_info_updated"

    table_elements: int | None
    increment_size: int | None
    table_csm: int | None


@dataclass(frozen=True, slots=True)
class LightTableInfoUpdated(Event):
    KIND: ClassVar[str] = "light_table_info_updated"

    table_elements: int | None
    increment_size: int | None
    table_csm: int | None


@dataclass(frozen=True, slots=True)
class BarrierTableInfoUpdated(Event):
    KIND: ClassVar[str] = "barrier_table_info_updated"

    table_elements: int | None
    increment_size: int | None
    table_csm: int | None


@dataclass(frozen=True, slots=True)
class LockTableInfoUpdated(Event):
    KIND: ClassVar[str] = "lock_table_info_updated"

    table_elements: int | None
    increment_size: int | None
    table_csm: int | None


# -------------------------
# CSM events
# -------------------------


@dataclass(frozen=True, slots=True)
class CsmSnapshotUpdated(Event):
    KIND: ClassVar[str] = "csm_snapshot_updated"

    snapshot: CsmSnapshot


@dataclass(frozen=True, slots=True)
class DomainCsmChanged(Event):
    KIND: ClassVar[str] = "domain_csm_changed"

    csm_domain: str = ""
    old: int | None = None
    new: int = 0

    @property
    @override
    def domain(self) -> str:
        return self.csm_domain or super().domain


@dataclass(frozen=True, slots=True)
class TableCsmChanged(Event):
    KIND: ClassVar[str] = "table_csm_changed"

    csm_domain: str = ""
    old: int | None = None
    new: int = 0

    @property
    @override
    def domain(self) -> str:
        return self.csm_domain or super().domain


# -------------------------
# Trouble / diagnostics
# -------------------------


@dataclass(frozen=True, slots=True)
class TroubleStatusUpdated(Event):
    KIND: ClassVar[str] = "trouble_status_updated"

    active: bool | None
    changed_fields: tuple[str, ...]


# -------------------------
# Panel version info
# -------------------------


@dataclass(frozen=True, slots=True)
class PanelVersionInfoUpdated(Event):
    KIND: ClassVar[str] = "panel_version_info_updated"

    changed_fields: tuple[str, ...]


# -------------------------
# API / protocol errors
# -------------------------


@dataclass(frozen=True, slots=True)
class ApiError(Event):
    KIND: ClassVar[str] = "api_error"

    error_code: int
    scope: str | None = None
    entity_id: int | None = None
    message: str | None = None


@dataclass(frozen=True, slots=True)
class AuthorizationRequiredEvent(Event):
    KIND: ClassVar[str] = "authorization_required"

    error_code: int
    scope: str | None = None
    entity_id: int | None = None
    message: str | None = None


@dataclass(frozen=True, slots=True)
class AuthenticateResult(Event):
    KIND: ClassVar[str] = "authenticate_result"

    success: bool
    error_code: int | None = None


@dataclass(frozen=True, slots=True)
class BootstrapCountsReady(Event):
    KIND: ClassVar[str] = "bootstrap_counts_ready"


@dataclass(frozen=True, slots=True)
class DispatchRoutingError(Event):
    KIND: ClassVar[str] = "dispatch_routing_error"

    code: str
    message: str
    keys: tuple[str, ...]
    severity: str  # "debug"|"info"|"warning"|"error"


# -------------------------
# Unknown/unhandled
# -------------------------


@dataclass(frozen=True, slots=True)
class UnknownMessage(Event):
    KIND: ClassVar[str] = "unknown_message"

    unhandled_route: RouteKey
    keys: tuple[str, ...]


# -------------------------
# Stamping helper (used by kernel.emit)
# -------------------------


def stamp_event(
    evt: Event,
    *,
    at: float,
    seq: int | None,
    classification: str,
    route: RouteKey,
    session_id: int | None,
) -> Event:
    """
    Replace the common header fields on an event.
    kernel.emit() uses this to make headers authoritative and consistent.
    """
    return replace(
        evt,
        at=at,
        seq=seq,
        classification=classification,
        route=route,
        session_id=session_id,
    )



"""
elke27_lib/features/bus_ios.py

Feature module: bus_ios

Responsibilities:
- Register inbound handlers for bus_io_dev.*
- Register outbound request builders for bus_io_dev.* routes
"""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from elke27_lib.kernel import E27Kernel

from elke27_lib.handlers.bus_ios import make_bus_ios_get_trouble_handler

ROUTE_BUS_IOS_GET_TROUBLE = ("bus_io_dev", "get_trouble")


def register(elk: E27Kernel) -> None:
    elk.register_handler(
        ROUTE_BUS_IOS_GET_TROUBLE,
        make_bus_ios_get_trouble_handler(elk.state, elk.emit, elk.now),
    )

    elk.register_request(
        ROUTE_BUS_IOS_GET_TROUBLE,
        build_bus_ios_get_trouble_payload,
    )


def build_bus_ios_get_trouble_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}



from __future__ import annotations

import json
import os
import socket
import time

import pytest

from elke27_lib.framing import DeframeState, deframe_feed
from elke27_lib.linking import (
    E27Identity,
    build_api_link_request,
    derive_pass_tempkey_with_cnonce,
    send_unframed_json,
    wait_for_discovery_nonce,
)
from elke27_lib.presentation import decrypt_api_link_response


def _env(name: str, default: str | None = None) -> str | None:
    value = os.environ.get(name, default)
    if value == "":
        return default
    return value


def _require_env(name: str) -> str:
    value = _env(name)
    if not value:
        pytest.skip(f"Missing {name} for live link test.")
    return value


@pytest.mark.integration
def test_link_raw_framed_and_deframed_buffers() -> None:
    host = _require_env("ELKE27_HOST")
    port = int(_env("ELKE27_PORT", "2101") or 2101)
    access_code = _require_env("ELKE27_ACCESS_CODE")
    passphrase = _require_env("ELKE27_PASSPHRASE")
    mn = _require_env("ELKE27_MN")
    sn = _require_env("ELKE27_SN")
    fwver = _require_env("ELKE27_FWVER")
    hwver = _require_env("ELKE27_HWVER")
    osver = _require_env("ELKE27_OSVER")

    timeout_s = float(_env("ELKE27_TIMEOUT_S", "5.0") or 5.0)

    client_identity = E27Identity(mn=mn, sn=sn, fwver=fwver, hwver=hwver, osver=osver)

    with socket.create_connection((host, port), timeout=timeout_s) as sock:
        nonce = wait_for_discovery_nonce(sock, timeout_s=timeout_s)
        cnonce = os.urandom(20).hex().lower()
        pass8, tempkey_hex = derive_pass_tempkey_with_cnonce(
            access_code=access_code,
            passphrase=passphrase,
            nonce=nonce,
            cnonce=cnonce,
            mn=mn,
            sn=sn,
        )

        req = build_api_link_request(
            seq=110,
            client_identity=client_identity,
            pass_hex8=pass8,
            cnonce_hex=cnonce,
        )
        send_unframed_json(sock, req)

        state = DeframeState()
        raw_chunks: list[bytes] = []
        frame_no_crc: bytes | None = None

        deadline = time.monotonic() + timeout_s
        while time.monotonic() < deadline and frame_no_crc is None:
            remaining = max(0.0, deadline - time.monotonic())
            sock.settimeout(min(0.5, remaining))
            try:
                chunk = sock.recv(4096)
            except TimeoutError:
                continue
            if not chunk:
                break
            raw_chunks.append(chunk)

            for res in deframe_feed(state, chunk):
                if res.ok and res.frame_no_crc:
                    frame_no_crc = res.frame_no_crc
                    break

        if frame_no_crc is None:
            pytest.fail("No framed response received within timeout.")

        framed_hex = b"".join(raw_chunks).hex()
        deframed_hex = frame_no_crc.hex()
        encrypted_hex = frame_no_crc[3:].hex()
        _ack, decrypted_bytes = decrypt_api_link_response(
            protocol_byte=frame_no_crc[0],
            ciphertext=frame_no_crc[3:],
            tempkey_hex=tempkey_hex,
        )
        decrypted_hex = decrypted_bytes.hex()
        try:
            decrypted_text = decrypted_bytes.decode("utf-8", errors="strict")
            decrypted_json = json.loads(decrypted_text)
        except (UnicodeDecodeError, json.JSONDecodeError):
            decrypted_json = None

        print(f"\nframed_hex={framed_hex}")
        print(f"deframed_hex={deframed_hex}")
        print(f"encrypted_hex={encrypted_hex}")
        print(f"decrypted_hex={decrypted_hex}")
        if decrypted_json is not None:
            print(f"decrypted_json={decrypted_json}")



"""Generated generator stubs and re-exports for all permission keys."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any, cast

from elke27_lib.permissions import ALL_PERMISSION_KEYS, canonical_generator_key

from . import (
    area,
    barrier,
    bus_ios,
    control,
    keypad,
    light,
    lock,
    log,
    network_param,
    output,
    rule,
    system,
    tstat,
    user,
    zone,
)

GeneratorFn = Callable[..., tuple[dict[str, object], tuple[str, str]]]


def _collect_generators(module: object) -> dict[str, GeneratorFn]:
    found: dict[str, GeneratorFn] = {}
    for name in dir(module):
        if not name.startswith("generator_"):
            continue
        fn = getattr(module, name)
        if callable(fn):
            key = canonical_generator_key(fn.__name__)
            found[key] = cast(GeneratorFn, fn)
    return found


def _stub_generator(key: str) -> GeneratorFn:
    def _stub(*_args: Any, **_kwargs: Any) -> tuple[dict[str, object], tuple[str, str]]:
        raise NotImplementedError(f"{key} not implemented in library yet")

    _stub.__name__ = f"generator_{key}"
    return _stub


_EXISTING: dict[str, GeneratorFn] = {}
for _module in (
    area,
    barrier,
    bus_ios,
    control,
    keypad,
    light,
    lock,
    log,
    network_param,
    output,
    rule,
    system,
    tstat,
    user,
    zone,
):
    _EXISTING.update(_collect_generators(_module))


GENERATORS: dict[str, GeneratorFn] = {}
for _key in sorted(ALL_PERMISSION_KEYS):
    fn = _EXISTING.get(_key) or _stub_generator(_key)
    GENERATORS[_key] = fn
    globals()[f"generator_{_key}"] = fn


__all__ = ["GENERATORS"]



"""
elke27_lib/handlers/tstat.py

Read/observe-only handlers for the "tstat" domain.
"""

from __future__ import annotations

import logging
from collections.abc import Callable, Mapping
from typing import Any, cast

from elke27_lib.dispatcher import DispatchContext, PagedBlock
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    ApiError,
    AuthorizationRequiredEvent,
    BootstrapCountsReady,
    CsmSnapshotUpdated,
    Event,
    TableCsmChanged,
    TstatConfiguredInventoryReady,
    TstatConfiguredUpdated,
    TstatStatusUpdated,
    TstatTableInfoUpdated,
)
from elke27_lib.states import PanelState, TstatState, update_csm_snapshot

EmitFn = Callable[[Event, DispatchContext], None]
NowFn = Callable[[], float]

LOG = logging.getLogger(__name__)


def _as_mapping(obj: object) -> Mapping[str, Any] | None:
    if isinstance(obj, Mapping):
        return cast(Mapping[str, Any], obj)
    return None


def _coerce_int(value: object) -> int | None:
    return value if isinstance(value, int) else None


def make_tstat_get_status_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("tstat","get_status").
    """

    def handler_tstat_get_status(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        tstat_obj = _as_mapping(msg.get("tstat"))
        if tstat_obj is None:
            return False

        payload = _as_mapping(tstat_obj.get("get_status"))
        if payload is None:
            return False

        error_code = payload.get("error_code")
        if isinstance(error_code, int) and error_code != 0:
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="tstat",
                    entity_id=_coerce_int(payload.get("tstat_id")),
                    message=None,
                ),
                ctx,
            )
            return True

        tstat_id = payload.get("tstat_id")
        if not isinstance(tstat_id, int) or tstat_id < 1:
            return False

        tstat = state.get_or_create_tstat(tstat_id)
        changed: set[str] = set()
        _apply_tstat_status_fields(tstat, payload, changed)
        tstat.last_update_at = now()
        state.panel.last_message_at = tstat.last_update_at

        emit(
            TstatStatusUpdated(
                kind=TstatStatusUpdated.KIND,
                at=UNSET_AT,
                seq=UNSET_SEQ,
                classification=UNSET_CLASSIFICATION,
                route=UNSET_ROUTE,
                session_id=UNSET_SESSION_ID,
                tstat_id=tstat_id,
                mode=tstat.mode,
                fan_mode=tstat.fan_mode,
                temperature=tstat.temperature,
            ),
            ctx,
        )
        return True

    return handler_tstat_get_status


def make_tstat_set_status_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("tstat","set_status").
    """

    def handler_tstat_set_status(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        tstat_obj = _as_mapping(msg.get("tstat"))
        if tstat_obj is None:
            return False

        payload = _as_mapping(tstat_obj.get("set_status"))
        if payload is None:
            return False

        error_code = payload.get("error_code")
        if isinstance(error_code, int) and error_code != 0:
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="tstat",
                    entity_id=_coerce_int(payload.get("tstat_id")),
                    message=None,
                ),
                ctx,
            )
            return True

        tstat_id = payload.get("tstat_id")
        if not isinstance(tstat_id, int) or tstat_id < 1:
            return False

        tstat = state.get_or_create_tstat(tstat_id)
        changed: set[str] = set()
        _apply_tstat_status_fields(tstat, payload, changed)
        tstat.last_update_at = now()
        state.panel.last_message_at = tstat.last_update_at

        emit(
            TstatStatusUpdated(
                kind=TstatStatusUpdated.KIND,
                at=UNSET_AT,
                seq=UNSET_SEQ,
                classification=UNSET_CLASSIFICATION,
                route=UNSET_ROUTE,
                session_id=UNSET_SESSION_ID,
                tstat_id=tstat_id,
                mode=tstat.mode,
                fan_mode=tstat.fan_mode,
                temperature=tstat.temperature,
            ),
            ctx,
        )
        return True

    return handler_tstat_set_status


def make_tstat_get_configured_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("tstat","get_configured").
    """

    def handler_tstat_get_configured(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        tstat_obj = _as_mapping(msg.get("tstat"))
        if tstat_obj is None:
            return False

        payload = _as_mapping(tstat_obj.get("get_configured"))
        if payload is None:
            return False

        error_code = payload.get("error_code", tstat_obj.get("error_code"))
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="tstat",
                        entity_id=None,
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="tstat",
                    entity_id=None,
                    message=None,
                ),
                ctx,
            )
            return True

        ids = _extract_configured_tstat_ids(payload)
        table_info = _as_mapping(state.table_info_by_domain.get("tstat"))
        if table_info is not None:
            max_id = table_info.get("table_elements")
            if isinstance(max_id, int) and max_id >= 1:
                ids = [tstat_id for tstat_id in ids if tstat_id <= max_id]

        inv = state.inventory
        inv.configured_tstats = set(ids)
        inv.configured_tstats_complete = True
        state.panel.last_message_at = now()

        emit(
            TstatConfiguredUpdated(
                kind=TstatConfiguredUpdated.KIND,
                at=UNSET_AT,
                seq=UNSET_SEQ,
                classification=UNSET_CLASSIFICATION,
                route=UNSET_ROUTE,
                session_id=UNSET_SESSION_ID,
                configured_ids=tuple(sorted(ids)),
            ),
            ctx,
        )
        emit(
            TstatConfiguredInventoryReady(
                kind=TstatConfiguredInventoryReady.KIND,
                at=UNSET_AT,
                seq=UNSET_SEQ,
                classification=UNSET_CLASSIFICATION,
                route=UNSET_ROUTE,
                session_id=UNSET_SESSION_ID,
            ),
            ctx,
        )
        return True

    return handler_tstat_get_configured


def make_tstat_configured_merge(_state: PanelState):
    def _merge(blocks: list[PagedBlock], block_count: int) -> Mapping[str, Any]:
        merged: set[int] = set()
        for block in blocks:
            for tstat_id in _extract_configured_tstat_ids(block.payload):
                merged.add(tstat_id)
        return {"tstats": sorted(merged), "block_count": block_count}

    return _merge


def make_tstat_get_attribs_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("tstat","get_attribs").
    """

    def handler_tstat_get_attribs(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        tstat_obj = _as_mapping(msg.get("tstat"))
        if tstat_obj is None:
            return False

        payload = _as_mapping(tstat_obj.get("get_attribs"))
        if payload is None:
            return False

        error_code = payload.get("error_code", tstat_obj.get("error_code"))
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="tstat",
                        entity_id=_coerce_int(payload.get("tstat_id")),
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="tstat",
                    entity_id=_coerce_int(payload.get("tstat_id")),
                    message=None,
                ),
                ctx,
            )
            return True

        tstat_id = payload.get("tstat_id")
        if not isinstance(tstat_id, int) or tstat_id < 1:
            return False

        tstat = state.get_or_create_tstat(tstat_id)
        changed: set[str] = set()
        _apply_tstat_attribs(tstat, payload, changed)
        tstat.last_update_at = now()
        state.panel.last_message_at = tstat.last_update_at
        return True

    return handler_tstat_get_attribs


def make_tstat_get_table_info_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("tstat","get_table_info").
    """

    def handler_tstat_get_table_info(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        tstat_obj = _as_mapping(msg.get("tstat"))
        if tstat_obj is None:
            return False

        payload = _as_mapping(tstat_obj.get("get_table_info"))
        if payload is None:
            payload = _as_mapping(tstat_obj.get("table_info"))
        if payload is None:
            return False

        error_code = payload.get("error_code")
        if isinstance(error_code, int) and error_code != 0:
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="tstat",
                    entity_id=None,
                    message=None,
                ),
                ctx,
            )
            return True

        table_info = dict(payload)
        state.table_info_by_domain["tstat"] = table_info
        state.panel.last_message_at = now()
        table_elements = _extract_int(payload, "table_elements")
        table_csm = _extract_table_csm(payload, domain="tstat")
        if table_csm is not None:
            old = state.table_csm_by_domain.get("tstat")
            if old != table_csm:
                state.table_csm_by_domain["tstat"] = table_csm
                emit(
                    TableCsmChanged(
                        kind=TableCsmChanged.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        csm_domain="tstat",
                        old=old,
                        new=table_csm,
                    ),
                    ctx,
                )
        if table_elements is not None:
            state.table_info_known.add("tstat")

        emit(
            TstatTableInfoUpdated(
                kind=TstatTableInfoUpdated.KIND,
                at=UNSET_AT,
                seq=UNSET_SEQ,
                classification=UNSET_CLASSIFICATION,
                route=UNSET_ROUTE,
                session_id=UNSET_SESSION_ID,
                table_elements=table_elements,
                increment_size=_extract_int(payload, "increment_size"),
                table_csm=table_csm,
            ),
            ctx,
        )

        snapshot = update_csm_snapshot(state)
        if snapshot is not None:
            emit(
                CsmSnapshotUpdated(
                    kind=CsmSnapshotUpdated.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    snapshot=snapshot,
                ),
                ctx,
            )

        if not state.bootstrap_counts_ready and {"area", "zone", "output", "tstat"}.issubset(
            state.table_info_known
        ):
            state.bootstrap_counts_ready = True
            emit(
                BootstrapCountsReady(
                    kind=BootstrapCountsReady.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                ),
                ctx,
            )
        return True

    return handler_tstat_get_table_info


def _apply_tstat_status_fields(
    tstat: TstatState, payload: Mapping[str, Any], changed: set[str]
) -> None:
    _maybe_set(tstat, "temperature", payload.get("temperature"), changed)
    _maybe_set(tstat, "cool_setpoint", payload.get("cool_setpoint"), changed)
    _maybe_set(tstat, "heat_setpoint", payload.get("heat_setpoint"), changed)
    _maybe_set(tstat, "mode", payload.get("mode"), changed)
    _maybe_set(tstat, "fan_mode", payload.get("fan_mode"), changed)
    _maybe_set(tstat, "humidity", payload.get("humidity"), changed)
    _maybe_set(tstat, "rssi", payload.get("rssi"), changed)

    battery = payload.get("battery level")
    if battery is None:
        battery = payload.get("battery_level")
    _maybe_set(tstat, "battery_level", battery, changed)

    prec = payload.get("prec")
    if isinstance(prec, list):
        prec_items = cast(list[object], prec)
        prec_values: list[int] = []
        all_ints = True
        for item in prec_items:
            if not isinstance(item, int):
                all_ints = False
                break
            prec_values.append(item)
        if all_ints:
            _maybe_set(tstat, "prec", prec_values, changed)

    for key, value in payload.items():
        if key in {
            "tstat_id",
            "error_code",
            "temperature",
            "cool_setpoint",
            "heat_setpoint",
            "mode",
            "fan_mode",
            "humidity",
            "rssi",
            "battery level",
            "battery_level",
            "prec",
        }:
            continue
        if tstat.fields.get(key) != value:
            tstat.fields[key] = value
            changed.add(key)


def _maybe_set(tstat: TstatState, attr: str, value: Any, changed: set[str]) -> None:
    if value is None:
        return
    if getattr(tstat, attr) != value:
        setattr(tstat, attr, value)
        changed.add(attr)


def _extract_int(payload: Mapping[str, Any], key: str) -> int | None:
    value = payload.get(key)
    return value if isinstance(value, int) else None


def _extract_table_csm(payload: Mapping[str, Any], *, domain: str) -> int | None:
    if "table_csm" not in payload:
        return None
    value = payload.get("table_csm")
    if isinstance(value, bool):
        LOG.warning("%s.get_table_info table_csm has invalid bool value.", domain)
        return None
    if isinstance(value, int):
        return value
    if isinstance(value, float) and value.is_integer():
        return int(value)
    if isinstance(value, str):
        text = value.strip()
        if text.isdigit():
            return int(text)
    LOG.warning("%s.get_table_info table_csm has non-int value %r.", domain, value)
    return None


def _extract_configured_tstat_ids(payload: Mapping[str, Any]) -> list[int]:
    keys = ("tstats", "tstat_ids", "configured_tstats", "configured_tstat_ids")
    for key in keys:
        value = payload.get(key)
        if isinstance(value, list):
            ids = [
                item for item in cast(list[object], value) if isinstance(item, int) and item >= 1
            ]
            return sorted(set(ids))
    return []


def _apply_tstat_attribs(tstat: TstatState, payload: Mapping[str, Any], changed: set[str]) -> None:
    name = payload.get("name")
    if isinstance(name, str):
        normalized_name = name.strip() or None
        if tstat.name != normalized_name:
            tstat.name = normalized_name
            changed.add("name")
    for key, value in payload.items():
        if key in {"tstat_id", "error_code", "name"}:
            continue
        if tstat.fields.get(key) != value:
            tstat.fields[key] = value
            changed.add(key)
